
#include <Streaming.h>
#include <STM32ADC.h>

STM32ADC adcV(ADC1);
STM32ADC adcI(ADC2);

uint8 adcPinV = PB0;
uint8 adcPinI = PB1;

byte  outputDrive = 0;    // Output Driver : High Byte of PORT B 
uint16 cycleIndex;           // position within the cycle : between 0 to 1023 , 
uint8  lastVoltageLookupIndex;
uint16  batteryVoltage = 26500;
int32 inputFactor = 1000;
uint32 nominalBatteryMicroVolts;
uint8 voltSteps;
uint16 quarterSteps;

volatile uint16 adcValueV;
volatile uint16 adcValueI;

HardwareTimer inverterTimer(4);

#define NOMINAL_VAC  240
#define NOMINAL_BATTERY_VOLTAGE 26500    // in milliVolts
#define TRANSFORMER_COUNT 4
#define STEPS 1024

uint16 Vtransformer[TRANSFORMER_COUNT];   // in milliVolts * 10 ie 23500 = 235.00 volts

uint16 sineLookup[STEPS];   
int16 outputVoltageLookup[3^TRANSFORMER_COUNT];   // in milliVolts * 10 ie 23500 = 235.00 volts
int8 transformerMask[3^TRANSFORMER_COUNT][TRANSFORMER_COUNT];
uint8 outputMaskTable[3^TRANSFORMER_COUNT];

void setup() 
{
  //slow way
  pinMode(PB8,OUTPUT);
  pinMode(PB9,OUTPUT);
  pinMode(PB10,OUTPUT);
  pinMode(PB11,OUTPUT);
  pinMode(PB12,OUTPUT);
  pinMode(PB13,OUTPUT);
  pinMode(PB14,OUTPUT);
  pinMode(PB15,OUTPUT);
  // quick way, the upper 8 bits all set to output, pushpull
  GPIOB->regs->CRH = 0x33333333;
  GPIOB->regs->ODR = 0;

  voltSteps = 3^TRANSFORMER_COUNT;
  quarterSteps = STEPS/4;
  nominalBatteryMicroVolts = NOMINAL_BATTERY_VOLTAGE * 1000;
  setTransformerVoltage();
  fill_TransformerMask();
  fill_outputDriveTable();
  fill_outputVoltageTable();
  fill_sineLookup();

//*******  ADC setup

  adcV.calibrate();
  adcV.setSampleRate(ADC_SMPR_239_5);
  adcV.attachInterrupt(adcConvertedInt, ADC_EOC);
  adcV.setPins(&adcPinV, 1);

  adcI.calibrate();
  adcI.setSampleRate(ADC_SMPR_239_5);
  //adcI.attachInterrupt(int_ADCI,, ADC2_EOC);
  adcI.setPins(&adcPinI, 1);

    // -- sets up the adc control registers
    //    for Dual ADC, simultaneous sample
  ADC1->regs->CR1 |= 0x60000;
  ADC1->regs->CR2 |= 0x100;
  
  //*******  Timer setup  ( 19.52777777777777777778 microseconds, 51209.10384068278805121 Hz)
  //---  output Frquency = 50.00889046941678520626 Hz
  //---  inverterTimer frequency = 51200Hz (50Hz x 1024 steps)
  //---  inverterTimer period = 19.53125 uSec
  //---  overflow = 72MHz / 51200Hz = 1406.25
  
  inverterTimer.setMode(TIMER_CH4, TIMER_OUTPUT_COMPARE);
  inverterTimer.pause();
  inverterTimer.setPrescaleFactor(1);
  inverterTimer.setCount(0);
  inverterTimer.setOverflow(1406);//11250
  inverterTimer.setCompare(TIMER_CH4, 1);  
  inverterTimer.attachCompare4Interrupt(timer_Handler);
  inverterTimer.refresh();

}



void timer_Handler()
{
  uint16 mask=0;
  static int8 cnt=0;
  byte portBlow = GPIOB->regs->ODR & 0x00ff;
  mask = outputDrive<<8 | portBlow;
  //--- write PORTB 
  GPIOB->regs->ODR = mask;

  cnt++;
  if (cnt>100)
  {
    cnt=0;
    adcV.startConversion();
  }
  //--- now work out the next driver value
  
  proc_nextOutputValue();
  
  //----  turn testPin1 high
  //gpio_write_bit(GPIOA, 4, HIGH);
  
}

void adcConvertedInt()
{
  static uint16 lastADC = 0;
  //gpio_write_bit(GPIOA, 4, LOW);
  //-- read adc 32bit data register
  uint32 data = ADC1->regs->DR;
  adcValueV = data & 0xFFFF;
  adcValueI = data >> 16;
}

void proc_nextOutputValue()
{
  cycleIndex ++;
  if (cycleIndex >= STEPS) 
  {
    cycleIndex = 0;
    proc_NewCycle();
  }
  int32 desiredVolts = sineLookup[cycleIndex];
  desiredVolts = desiredVolts * inputFactor + 500;
  desiredVolts = (int32) (desiredVolts / 1000 );
}

void proc_NewCycle()
{
  //inputFactor = (int32) ( batteryVoltage * 100u) ;
  //inputFactor = inputFactor / NOMINAL_BATTERY_VOLTAGE;
  inputFactor = nominalBatteryMicroVolts / batteryVoltage;
}

uint8 findMask(uint currentStep,int32 matchVolts)
{
  int16 index;
  
  if (currentStep < quarterSteps )
  {
    index = findMaskUp(lastVoltageLookupIndex,matchVolts);
    if (index<0)
      {
        index = findMaskUp(0,matchVolts);
      }
  }

  if (currentStep >= quarterSteps && currentStep < 2*quarterSteps)
  {
    index = findMaskDn(lastVoltageLookupIndex,matchVolts);
    if (index<0)
      {
        index = findMaskDn(0,matchVolts);
      }
  }

  if (currentStep >= 2*quarterSteps && currentStep < 3*quarterSteps )
  {
    index = findMaskUp(lastVoltageLookupIndex,matchVolts);
    if (index<0)
      {
        index = findMaskUp(0,matchVolts);
      }
  }

  if (currentStep >= 3*quarterSteps && currentStep < 4*quarterSteps)
  {
    index = findMaskDn(lastVoltageLookupIndex,matchVolts);
    if (index<0)
      {
        index = findMaskDn(0,matchVolts);
      }
  }
  
  if (lastVoltageLookupIndex != index)
  {
    lastVoltageLookupIndex = index;
    outputDrive = outputMaskTable[index];
  }
}



int16 findMaskUp(uint8 lookupindex, int32 matchVolts)
{
  for (int i=lookupindex;i<voltSteps;i++)
  {
    if (matchVolts < outputVoltageLookup[i])
    {
      if (matchVolts > outputVoltageLookup[i])
      {
        return (i);
      }
    }
  }
  return -1;
}

int16 findMaskDn(uint8 lookupindex, int32 matchVolts)
{
  for (int i=lookupindex;i<voltSteps;i++)
  {
    if (matchVolts > outputVoltageLookup[i])
    {
      if (matchVolts < outputVoltageLookup[i])
      {
        return (i);
      }
    }
  }
  return -1;
}

void setTransformerVoltage()
{
  float Vpeak =  NOMINAL_VAC * sqrt(2) * 100;
  for (int i=0;i<TRANSFORMER_COUNT;i++)
  {
    int x = TRANSFORMER_COUNT - i - 1;
    float F = (3^x) / (3^TRANSFORMER_COUNT) ;
    Vtransformer[x] = (uint16) (Vpeak * F);
  }
}

/*--------------------------------------------------------------------------------
 * 
 *  Fills  array transformeMask with an incremented ternary (base 3) mask
 *  
 *  in relation to the transformer  0 = -1, 1 = 0, 2 = +1 
 * 
 --------------------------------------------------------------------------------*/

void fill_TransformerMask()
{
  for (int i=0;i<TRANSFORMER_COUNT;i++)
  {
    transformerMask[0][i] = 0;
  }
  for (int i=1;i<3^TRANSFORMER_COUNT;i++)
  {
    for (int j=0;j<TRANSFORMER_COUNT;j++)
    {
      if (j==0)
      {
        transformerMask[i][j] = 0;
        if (transformerMask[i-1][j] < 2)  transformerMask[i][j] = transformerMask[i-1][j] +1;
      }
      if (j>0 && transformerMask[i-1][j-1] > transformerMask[i][j-1])
      {
        transformerMask[i][j] = 0;
        if (transformerMask[i-1][j] < 2) transformerMask[i][j] = transformerMask[i-1][j] +1;
      }
    }
  }
}

void fill_outputDriveTable()
{
  for (int i=1;i<3^TRANSFORMER_COUNT;i++)
  {
    outputMaskTable[i] = 0 ;
    for (int j=0;j<TRANSFORMER_COUNT;j++)
    {
      int mult = 4^j;
      outputMaskTable[i] += transformerMask[i][j] * mult;
    }
  }
}

void fill_outputVoltageTable()
{
  for (int i=1;i<3^TRANSFORMER_COUNT;i++)
  {
    int16 sumVolts = 0;
    for (int j=0;j<TRANSFORMER_COUNT;j++)
    {
      sumVolts += transformerMask[i-1][j] * Vtransformer[j];
    }
    outputVoltageLookup[i] = sumVolts;
  }
}


void make_outputVoltageTable_junk()
{
  uint16 v[TRANSFORMER_COUNT];
  int8 Tstates[3^TRANSFORMER_COUNT][TRANSFORMER_COUNT];
  uint8 ternaryValue[TRANSFORMER_COUNT];
  for (int i=TRANSFORMER_COUNT-1;i>=0;i--)
  {
    ternaryValue[i] = 3^i;
  }
  
  for (int i=0;i<3^TRANSFORMER_COUNT;i++)
  {
    int junk;
    for (int j=0;j<TRANSFORMER_COUNT;j++)
    {
      int a = int (i/ternaryValue[j]);
      junk = i-a*ternaryValue[j];
    }
  }
}

void fill_sineLookup()
{
  double Vpeak =  NOMINAL_VAC * sqrt(2) * 100;
  for (int i=0;i< STEPS; i++)
  {
    double rad = 2 * PI / STEPS * i;
    double volts = sin(rad) * Vpeak;
    sineLookup[i] = int(volts * 100 + 0.5);
  }
}

uint16 MAKE_WORD( const byte Byte_hi, const byte Byte_lo)
{
     return   (( Byte_hi << 8  ) | Byte_lo & 0x00FF );
}

void loop() {
  // put your main code here, to run repeatedly:

}
